# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os, sys, re
sys.path.insert(0, os.path.abspath('../'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'PyRemoteData'
copyright = '2024, Asger Svenning'
author = 'Asger Svenning'
release = [re.search(r'version\s*=\s*\"([^\"]*)\"', line).group(1) for line in open('../pyproject.toml') if 'version' in line][0]

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode", 
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon"
]
napoleon_google_docstring = True
napoleon_numpy_docstring = False

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_material'
html_css_files = ['custom.css']
html_static_path = ['_static']
html_baseurl = '_build/html/'

html_theme_options = {

    # Set the name of the project to appear in the navigation.
    'nav_title': 'PyRemoteData',

    # # Specify a base_url used to generate sitemap.xml. If not
    # # specified, then no sitemap will be built.
    # 'base_url': 'https://project.github.io/project',

    # # Set the color and the accent color
    # 'color_primary': 'blue',
    # 'color_accent': 'light-blue',

    # Set the repo location to get a badge with stats
    'repo_url': 'https://github.com/asgersvenning/pyremotedata',
    'repo_name': 'pyremotedata',

    # Visible levels of the global TOC; -1 means unlimited
    'globaltoc_depth': 3,
    # If False, expand all TOC entries
    'globaltoc_collapse': False,
    # If True, show hidden TOC entries
    'globaltoc_includehidden': False,
}

intersphinx_mapping = {
    'pyremotedata': ("https://asgersvenning.github.io/pyremotedata/" + html_baseurl, None),
    'python': ('https://docs.python.org/3', None),
    'torch': ('https://pytorch.org/docs/stable/', None),
}

### Custom code to remove custom Sphinx comments from docstrings

import re

def preprocess_docstring(app, what, name, obj, options, lines):
    START_TAG = ".. <Sphinx comment"
    END_TAG = ".. Sphinx comment>"

    docstring = '\n'.join(lines)

    # Find all start and end tag positions
    start_positions = [match.start() for match in re.finditer(START_TAG, docstring)]
    end_positions = [match.start() for match in re.finditer(END_TAG, docstring)]

    comment_blocks = []

    string_index = 0
    while string_index < len(docstring) and len(start_positions) > 0:
        # Get the first unused start tag after the current string index
        for i, start_position in enumerate(start_positions):
            if start_position >= string_index:
                matched_start = i
                break
        else:
            matched_start = None
        
        # If there are no start tags after the current string index, then we are done
        if matched_start is None:
            break

        # Get the position of the first start tag
        first_start = start_positions.pop(matched_start)
        # And remove all start tags that are before the first start tag (we don't need to consider them anymore)
        start_positions = start_positions[matched_start:]

        # Find the first end tag that is after the start tag
        for i, end_position in enumerate(end_positions):
            if end_position > first_start:
                matched_end = i
                break
        else:
            matched_end = None
        
        # If there are no end tags after the start tag, then we insert an implicit end tag at the end of the docstring
        if matched_end is None:
            following_end = len(docstring)
        # Otherwise, we use the matched end tag
        else:
            following_end = end_positions.pop(matched_end)
            # And remove all end tags that are before the start tag (we don't need to consider them anymore)
            end_positions = end_positions[matched_end:]
        
        # Save the comment block start and end positions
        comment_blocks += [(first_start, following_end)]

        # Move the string index to the end of the comment block
        string_index = following_end + len(END_TAG)
    
    # Remove the comment blocks from the docstring
    for start, end in reversed(comment_blocks):
        docstring = docstring[:start] + docstring[end+len(END_TAG):]
    
    lines[:] = docstring.split('\n')
    
def setup(app):
    app.connect('autodoc-process-docstring', preprocess_docstring)


#################################################################################
#####                                                                       #####
##### OBS: I strongly discourage running this test on a production machine. #####
#####                                                                       #####
#################################################################################


## PSA: This test has external dependencies 'wrapt_timeout_decorator' and 'unittest' ##
import wrapt_timeout_decorator
import unittest
from unittest.mock import patch
import os, time, uuid

from pyremotedata import module_logger

class TestImplicitMount(unittest.TestCase):
    @wrapt_timeout_decorator.timeout(10)
    def test_implicit_mount(self):
        module_logger.info("Running basic functionality test.")
        with patch.dict('os.environ', {
            'PYREMOTEDATA_REMOTE_USERNAME': 'foo',
            'PYREMOTEDATA_REMOTE_URI': '0.0.0.0',
            'PYREMOTEDATA_REMOTE_DIRECTORY': 'upload',
            'PYREMOTEDATA_AUTO': 'yes'
        }):
            # Import the module
            from pyremotedata.implicit_mount import IOHandler
            # Open the connection
            with IOHandler() as handler:
                # Run some commands
                module_logger.info(handler.pwd())
                module_logger.info(handler.ls())
            
            from pyremotedata.config import remove_config
            remove_config()
        module_logger.info("Basic functionality test passed.")
    
    @wrapt_timeout_decorator.timeout(30)
    def test_error_assignment(self):
        module_logger.info("Running stress test on correct error assignment.")
        with patch.dict('os.environ', {
            'PYREMOTEDATA_REMOTE_USERNAME': 'foo',
            'PYREMOTEDATA_REMOTE_URI': '0.0.0.0',
            'PYREMOTEDATA_REMOTE_DIRECTORY': 'upload',
            'PYREMOTEDATA_AUTO': 'yes'
        }):
            n_exp_err = 1000
            # Import the module
            from pyremotedata.implicit_mount import IOHandler
            # Open the connection
            with IOHandler() as io:
                module_logger.info("Output before stress test:")
                module_logger.info(io.ls())
                # Interleave valid and invalid commands
                n_err = 0
                for _ in range(n_exp_err):
                    try:
                        print(io.cd(str(uuid.uuid4())))
                    except:
                        n_err += 1
                        pass
                    io.ls()
                module_logger.info("Output after stress test:")
                module_logger.info(io.ls())
            
            assert n_err == n_exp_err, f"Invalid number of errors generated: {n_err}, expected: {n_exp_err}"
            
            from pyremotedata.config import remove_config
            remove_config()
        module_logger.info("Error assignment test passed.")

class TestUploadDownload(unittest.TestCase):
    @wrapt_timeout_decorator.timeout(25)
    def test_upload_download(self):
        module_logger.info("Running upload/download test.")
        with patch.dict('os.environ', {
            'PYREMOTEDATA_REMOTE_USERNAME': 'foo',
            'PYREMOTEDATA_REMOTE_URI': '0.0.0.0',
            'PYREMOTEDATA_REMOTE_DIRECTORY': 'upload',
            'PYREMOTEDATA_AUTO': 'yes'
        }):
            # Import the module
            from pyremotedata.implicit_mount import IOHandler
            # Open the connection
            with IOHandler() as handler:
                module_logger.info(handler.pwd())
                # Upload a test file to the mock SFTP server
                test_file_size = 10 # MB
                n_rep = 10
                generate_test_file_command = f"bash -c 'openssl rand -out {os.path.join(handler.local_dir, 'localfile.txt')} -base64 {int(test_file_size * (10**6) * 3/4)}'"
                module_logger.info(f'Generating test file with command: {generate_test_file_command}')
                os.system(generate_test_file_command)
                start_upload = time.time()
                upload_result = handler.put("localfile.txt", "testfile.txt", execute=False)
                upload_result = handler.execute_command(f'repeat -c {n_rep} -d 0.01 "rm -f testfile.txt && {upload_result}"')
                end_upload = time.time()
                # Download the test file from the mock SFTP server
                start_download = time.time()
                download_result = handler.download("testfile.txt", "testfile.txt", execute=False)
                download_result = handler.execute_command(f'repeat -c {n_rep} -d 0.01 "(!rm -f testfile.txt) && {download_result}"')
                end_download = time.time()
                # Get the local directory where the file should be downloaded to
                local_directory = handler.local_dir
                # Sanity checks
                local_file_exists = os.path.exists(os.path.join(local_directory, 'testfile.txt'))
                local_file_size = os.path.getsize(os.path.join(local_directory, 'testfile.txt')) / 10**6
                if not local_file_exists:
                    raise RuntimeError("Something went wrong with the download. The file does not exist locally.")
                # Enforce a +/- 10% tolerance on the file size
                if not (local_file_size > (0.9 * test_file_size) and local_file_size < (1.1 * test_file_size)):
                    raise RuntimeError(f"Something went wrong with the download. The file size (~{local_file_size:.2f} MB) is not correct.")
            
                # Cleanup
                os.remove(os.path.join(handler.local_dir, "localfile.txt"))
                os.remove(os.path.join(handler.local_dir, "testfile.txt"))
            
            from pyremotedata.config import remove_config
            remove_config()

        # Calculate results
        upload_time = (end_upload - start_upload) / n_rep - 0.01
        download_time = (end_download - start_download) / n_rep - 0.01
        upload_speed, download_speed = 100/upload_time, 100/download_time
        module_logger.info(f'Upload/download test passed with upload {upload_speed:.1f} MB/s and download {download_speed:.1f} MB/s.') 

if __name__ == "__main__":
    unittest.main()


from pyremotedata.implicit_mount import *
import time
import tempfile
import os

test_size = 10**9

def get_mock_file(dir, size):
    """
    Creates a mock file of size `size` (bytes) in directory `dir`
    At the moment the function merely assumes that the file "sample.txt" exists in the directory of this script.
    it can be created with "openssl rand -out sample.txt -base64 $(( 2**30 * 3/4 ))" (takes a couple of seconds)
    """
    return "sample.txt"

def perf_test():
    upload_dir_obj = tempfile.TemporaryDirectory()
    download_dir_obj = tempfile.TemporaryDirectory()

    upload_dir = upload_dir_obj.name
    download_dir = download_dir_obj.name

    mock_file = get_mock_file(upload_dir, size=test_size)

    mount = ImplicitMount()
    mount.mount()
    mount.cd("testing")
    # Set the local working directory to the directory which contains this script
    mount.lcd(os.path.dirname(os.path.abspath(__file__)))

    start_upload = time.time()
    mount.put(mock_file)
    end_upload = time.time()

    start_download = time.time()
    mount.pget(mock_file, download_dir)
    end_download = time.time()

    start_delete = time.time()
    mount.execute_command("rm {}".format(mock_file))
    end_delete = time.time()

    upload_time = end_upload - start_upload
    download_time = end_download - start_download
    delete_time = end_delete - start_delete

    upload_speed, download_speed, delete_speed = test_size/upload_time, test_size/download_time, test_size/delete_time

    print("Upload speed: {} MB/s".format(upload_speed/10**6))
    print("Download speed: {} MB/s".format(download_speed/10**6))
    print("Delete speed: {} MB/s".format(delete_speed/10**6))

    mount.unmount()

    upload_dir_obj.cleanup()
    download_dir_obj.cleanup()

perf_test()


import unittest
from unittest.mock import patch

from pyremotedata import module_logger

class TestConfig(unittest.TestCase):
    def test_config(self):
        module_logger.info("Running config test.")
        with patch.dict('os.environ', {
            'PYREMOTEDATA_REMOTE_USERNAME': 'test',
            'PYREMOTEDATA_REMOTE_URI': 'test.com',
            'PYREMOTEDATA_REMOTE_DIRECTORY': 'test',
            'PYREMOTEDATA_AUTO': 'yes'
        }):
            import pyremotedata.config as config
            c = config.get_config()
            module_logger.info(c)
            assert c is not None
            # Cleanup
            config.remove_config()
        module_logger.info("Config test passed.")

if __name__ == "__main__":
    unittest.main()


from pyremotedata.implicit_mount import IOHandler
from tqdm import tqdm as TQDM

with IOHandler() as io:
    for _ in TQDM(range(1000), desc="Creating errors...", unit="err"):
        try:
            print(io.cd("timelapse"))
        except:
            pass
        io.ls()
    print(io.ls())